import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects import base_character, base_effect
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, argtypes


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "burn",
        help="Burns vertically in the output area.",
        description="burn | Burn the output area.",
        epilog="Example: terminaltexteffects burn -a 0.003 --flame-color ff9600 --burned-color 252525",
    )
    effect_parser.set_defaults(effect_class=BurnEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.003,
        help="Time between animation steps. Defaults to 0.03 seconds.",
    )
    effect_parser.add_argument(
        "--burned-color",
        type=argtypes.valid_color,
        default="252525",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color faded toward as blocks burn. Defaults to 252525",
    )
    effect_parser.add_argument(
        "--flame-color",
        type=argtypes.valid_color,
        default="ff9600",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the flame. Defaults to 0",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the final character. Will leave as system default if not provided.",
    )


class BurnEffect(base_effect.Effect):
    """Effect that burns up the screen."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        super().__init__(terminal, args)

    def prepare_data(self) -> None:
        """Prepares the data for the effect by building the burn animation and organizing the data into columns."""
        vertical_build_order = [
            ".",
            "▖",
            "▄",
            "▙",
            "█",
            "▜",
            "▀",
        ]
        fire_gradient = graphics.gradient("ffffff", self.args.flame_color, 12)
        burned_gradient = graphics.gradient(self.args.flame_color, self.args.burned_color, 7)
        groups = self.input_by_column()
        for column in groups.values():
            column.reverse()

        def groups_remaining(rows) -> bool:
            return any(row for row in rows.values())

        while groups_remaining(groups):
            keys = [key for key in groups.keys() if groups[key]]
            next_char = groups[random.choice(keys)].pop(0)
            next_char.is_active = False

            g_start = 0
            for _, block in enumerate(vertical_build_order[:5]):
                for color in fire_gradient[g_start : g_start + 3]:
                    next_char.animator.add_effect_to_scene("construct", block, color, duration=30)
                g_start += 2

            g_start = 0
            for _, block in enumerate(vertical_build_order[4:]):
                for color in burned_gradient[g_start : g_start + 3]:
                    next_char.animator.add_effect_to_scene("construct", block, color, duration=30)
                g_start += 2

            next_char.animator.add_effect_to_scene(
                "construct", next_char.input_symbol, self.args.final_color, duration=1
            )
            next_char.animator.activate_scene("construct")
            self.pending_chars.append(next_char)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        while self.pending_chars or self.animating_chars:
            if self.pending_chars:
                next_char = self.pending_chars.pop(0)
                next_char.is_active = True
                self.animating_chars.append(next_char)

            self.animate_chars()

            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animator.is_active_scene_complete()
            ]
            self.terminal.print()

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation."""
        for animating_char in self.animating_chars:
            animating_char.animator.step_animation()



import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects import base_effect
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "randomsequence",
        help="Prints the input data in a random sequence.",
        description="randomsequence | Prints the input data in a random sequence.",
        epilog="Example: terminaltexteffects randomsequence -a 0.01",
    )
    effect_parser.set_defaults(effect_class=RandomSequence)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.003,
        help="Time to sleep between animation steps. Defaults to 0.01 seconds.",
    )
    effect_parser.add_argument(
        "--fade-startcolor",
        type=argtypes.valid_color,
        default="000000",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color to start the fade-in. Defaults to black.",
    )
    effect_parser.add_argument(
        "--fade-endcolor",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color to fade towards. Defaults to white",
    )
    effect_parser.add_argument(
        "--fade-duration",
        type=argtypes.valid_duration,
        default=5,
        help="Duration the fade effect, in frames. Defaults to 5.",
    )


class RandomSequence(base_effect.Effect):
    """Prints the input data in a random sequence."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        """Initializes the effect.

        Args:
            terminal (Terminal): Terminal object.
            args (argparse.Namespace): Arguments from argparse.
        """
        super().__init__(terminal, args)
        self.fade_in = graphics.gradient(args.fade_startcolor, args.fade_endcolor, 10)

    def prepare_data(self) -> None:
        for character in self.terminal.characters:
            character.is_active = False
            for color in self.fade_in:
                character.animator.add_effect_to_scene("fade", character.input_symbol, color, self.args.fade_duration)
            character.animator.activate_scene("fade")
            self.pending_chars.append(character)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        random.shuffle(self.pending_chars)
        while self.pending_chars or self.animating_chars:
            if self.pending_chars:
                self.next_char = self.pending_chars.pop()
                self.next_char.is_active = True
                self.animating_chars.append(self.next_char)
            self.animate_chars()
            self.terminal.print()
            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animator.is_active_scene_complete()
            ]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation."""
        for animating_char in self.animating_chars:
            animating_char.animator.step_animation()



"""EffectCharacter class and supporting classes to initialize and manage the state of a single character from the input data."""

import typing
from enum import Enum, auto
from terminaltexteffects.utils import graphics, motion

if typing.TYPE_CHECKING:
    from terminaltexteffects.utils.terminal import Terminal


class EventHandler:
    """Handles events for the EffectCharacter."""

    def __init__(self, character: "EffectCharacter"):
        self.character = character
        self.registered_events: dict[tuple[EventHandler.Event, str], list[tuple[EventHandler.Action, str]]] = {}

    class Event(Enum):
        """Events that can be registered with the EventHandler."""

        WAYPOINT_REACHED = auto()
        "A waypoint has been reached."
        SCENE_COMPLETE = auto()
        "An animation scene has completed."

    class Action(Enum):
        """Actions that can be taken when an event is triggered."""

        ACTIVATE_WAYPOINT = auto()
        "Activates a waypoint. The action target is the waypoint ID."
        ACTIVATE_SCENE = auto()
        "Activates an animation scene. The action target is the scene ID."

    def register_event(self, event: Event, caller: str, action: Action, action_target: str) -> None:
        """Registers an event to be handled by the EventHandler.

        Args:
            event (Event): The event to register.
            caller (str): The ID of the caller.
            action (Action): The action to take when the event is triggered.
            action_target (str): The ID of the action target.
        """
        new_event = (event, caller)
        new_action = (action, action_target)
        if new_event not in self.registered_events:
            self.registered_events[new_event] = list()
        self.registered_events[new_event].append(new_action)

    def handle_event(self, event: Event, caller: str) -> None:
        """Handles an event.

        Args:
            event (Event): An event to handle. If the event is not registered, nothing happens.
            caller (str): The ID of the caller.
        """
        action_map = {
            EventHandler.Action.ACTIVATE_WAYPOINT: self.character.motion.activate_waypoint,
            EventHandler.Action.ACTIVATE_SCENE: self.character.animator.activate_scene,
        }

        if (event, caller) not in self.registered_events:
            return
        for event_action in self.registered_events[(event, caller)]:
            action, action_target = event_action
            action_map[action](action_target)


class EffectCharacter:
    """A single character from the input data. Contains the state of the character.

    An EffectCharacter object contains the symbol, animation units, graphical modes, waypoints, and coordinates for a single
    character from the input data. The EffectCharacter object is used by the Effect class to animate the character.

    Attributes:
        symbol (str): the current symbol used in place of the character.
        is_active (bool): active characters are printed to the terminal.
        input_symbol (str): the symbol for the character in the input data.
    """

    def __init__(self, symbol: str, input_column: int, input_row: int, terminal: "Terminal"):
        """Initializes the EffectCharacter class.

        Args:
            symbol (str): the character symbol.
            input_column (int): the final column position of the character.
            input_row (int): the final row position of the character.
        """
        self.input_symbol: str = symbol
        "The symbol for the character in the input data."
        self.input_coord: motion.Coord = motion.Coord(input_column, input_row)
        "The coordinate of the character in the input data."
        self.symbol: str = symbol
        "The current symbol for the character, determined by the animation units."
        self.terminal: Terminal = terminal
        self.animator: graphics.Animator = graphics.Animator(self)
        self.motion: motion.Motion = motion.Motion(self)
        self.event_handler: EventHandler = EventHandler(self)
        self.is_active: bool = False
        "Active characters are printed to the terminal."



"""Creates a rain effect where characters fall from the top of the terminal."""

import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.utils import graphics
from terminaltexteffects import base_character, base_effect
from terminaltexteffects.utils.terminal import Terminal


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "rain",
        formatter_class=argtypes.CustomFormatter,
        help="Rain characters from the top of the output area.",
        description="rain | Rain characters from the top of the output area.",
        epilog=f"""{argtypes.EASING_EPILOG}
        
Example: terminaltexteffects rain -a 0.01 --rain-colors 39 45 51 21""",
    )
    effect_parser.set_defaults(effect_class=RainEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time, in seconds, between animation steps.",
    )
    effect_parser.add_argument(
        "--rain-colors",
        type=argtypes.valid_color,
        nargs="*",
        default=0,
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="List of colors for the rain drops. Colors are randomly chosen from the list.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the final character.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.5,
        metavar="(float > 0)",
        help="Falling speed of the rain drops. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="IN_QUART",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement.",
    )


class RainEffect(base_effect.Effect):
    """Creates a rain effect where characters fall from the top of the output area."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        super().__init__(terminal, args)
        self.group_by_row: dict[int, list[base_character.EffectCharacter | None]] = {}

    def prepare_data(self) -> None:
        """Prepares the data for the effect by setting all characters y position to the input height and sorting by target y."""
        if self.args.rain_colors:
            rain_colors = self.args.rain_colors
        else:
            rain_colors = [39, 45, 51, 21, 117, 159]
        rain_characters = ["o", ".", ",", "*", "|"]

        for character in self.terminal.characters:
            raindrop_color = random.choice(rain_colors)
            character.animator.add_effect_to_scene("rain", random.choice(rain_characters), raindrop_color, 1)
            raindrop_gradient = graphics.gradient(raindrop_color, self.args.final_color, 7)
            for color in raindrop_gradient:
                character.animator.add_effect_to_scene("fade", character.input_symbol, color, 5)
            character.animator.activate_scene("rain")
            character.motion.set_coordinate(character.input_coord.column, self.terminal.output_area.top)
            character.motion.new_waypoint(
                "input_coord",
                character.input_coord.column,
                character.input_coord.row,
                speed=self.args.movement_speed,
                ease=self.args.easing,
            )
            character.event_handler.register_event(
                character.event_handler.Event.WAYPOINT_REACHED,
                "input_coord",
                character.event_handler.Action.ACTIVATE_SCENE,
                "fade",
            )
            character.motion.activate_waypoint("input_coord")
            self.pending_chars.append(character)
        for character in sorted(self.pending_chars, key=lambda c: c.input_coord.row):
            if character.input_coord.row not in self.group_by_row:
                self.group_by_row[character.input_coord.row] = []
            self.group_by_row[character.input_coord.row].append(character)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        self.pending_chars.clear()
        self.terminal.print()
        while self.group_by_row or self.animating_chars or self.pending_chars:
            if not self.pending_chars and self.group_by_row:
                self.pending_chars.extend(self.group_by_row.pop(min(self.group_by_row.keys())))  # type: ignore
            if self.pending_chars:
                for _ in range(random.randint(1, 3)):
                    if self.pending_chars:
                        next_character = self.pending_chars.pop(random.randint(0, len(self.pending_chars) - 1))
                        next_character.is_active = True
                        self.animating_chars.append(next_character)

                    else:
                        break
            self.animate_chars()
            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animator.is_active_scene_complete()
                or not animating_char.motion.movement_complete()
            ]
            self.terminal.print()

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and getting the next symbol from the animator."""
        for animating_char in self.animating_chars:
            animating_char.animator.step_animation()
            animating_char.motion.move()



import argparse

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects import base_effect
from terminaltexteffects.utils.terminal import Terminal


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "rowmerge",
        formatter_class=argtypes.CustomFormatter,
        help="Merges rows of characters.",
        description="rowmerge | Merges rows of characters.",
        epilog=f"""{argtypes.EASING_EPILOG}        
        
Example: terminaltexteffects rowmerge -a 0.01""",
    )
    effect_parser.set_defaults(effect_class=RowMergeEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time, in seconds, to sleep between animation steps.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.5,
        metavar="(float > 0)",
        help="Movement speed of the characters. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="OUT_QUAD",
        type=argtypes.valid_ease,
        help="Easing function to use for row movement.",
    )


class RowMergeEffect(base_effect.Effect):
    """Effect that merges rows."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        super().__init__(terminal, args)

    def prepare_data(self) -> None:
        """Prepares the data for the effect by setting every other row to start at the opposite
        side of the terminal and reverse the order."""

        self.rows = []
        for row_index, row in self.input_by_row().items():
            if row_index % 2 == 0:
                row = row[::-1]
                column = self.terminal.output_area.left
            else:
                column = self.terminal.output_area.right
            for character in row:
                character.is_active = False
                character.motion.set_coordinate(column, character.input_coord.row)
                character.motion.new_waypoint(
                    "input_coord",
                    character.input_coord.column,
                    character.input_coord.row,
                    speed=self.args.movement_speed,
                    ease=self.args.easing,
                )
                character.motion.activate_waypoint("input_coord")
            self.rows.append(row)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        while self.rows or self.animating_chars:
            self.animate_chars()
            for row in self.rows:
                if row:
                    next_character = row.pop(0)
                    next_character.is_active = True
                    self.animating_chars.append(next_character)
            self.rows = [row for row in self.rows if row]

            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.motion.movement_complete()
            ]
            self.terminal.print()

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and printing the characters to the terminal."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()



import argparse

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects import base_effect
from terminaltexteffects.utils.terminal import Terminal


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "expand",
        formatter_class=argtypes.CustomFormatter,
        help="Expands the text from a single point.",
        description="expand | Expands the text from a single point.",
        epilog=f"""{argtypes.EASING_EPILOG}
        
Example: terminaltexteffects expand -a 0.01""",
    )
    effect_parser.set_defaults(effect_class=ExpandEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time, in seconds, between animation steps.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.5,
        metavar="(float > 0)",
        help="Movement speed of the characters. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="IN_OUT_QUART",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement.",
    )


class ExpandEffect(base_effect.Effect):
    """Effect that draws the characters expanding from a single point."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        super().__init__(terminal, args)

    def prepare_data(self) -> None:
        """Prepares the data for the effect by starting all of the characters from a point in the middle of the input data."""

        for character in self.terminal.characters:
            character.motion.set_coordinate(self.terminal.output_area.right // 2, self.terminal.output_area.top // 2)
            character.motion.new_waypoint(
                "input_coord",
                character.input_coord.column,
                character.input_coord.row,
                speed=self.args.movement_speed,
                ease=self.args.easing,
            )
            character.is_active = True
            character.motion.activate_waypoint("input_coord")
            self.animating_chars.append(character)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        self.terminal.print()
        while self.animating_chars:
            self.animate_chars()
            self.animating_chars = [
                animating for animating in self.animating_chars if not animating.motion.movement_complete()
            ]
            self.terminal.print()

    def animate_chars(self) -> None:
        for animating_char in self.animating_chars:
            animating_char.motion.move()



"""A module for managing the terminal state and output."""
import shutil
import sys
import time
import argparse
from dataclasses import dataclass

from terminaltexteffects.base_character import EffectCharacter
from terminaltexteffects.utils import ansitools


@dataclass
class OutputArea:
    """A class for storing the output area of an effect.

    Args:
        top (int): top row of the output area
        right (int): right column of the output area
        bottom (int): bottom row of the output area. Defaults to 1.
        left (int): left column of the output area. Defaults to 1.

    """

    top: int
    right: int
    bottom: int = 1
    left: int = 1


class Terminal:
    """A class for managing the terminal state and output."""

    def __init__(self, input_data: str, args: argparse.Namespace):
        self.input_data = input_data
        self.width, self.height = self._get_terminal_dimensions()
        self.characters = self._decompose_input(args.xterm_colors, args.no_color)
        self.input_width = max([character.input_coord.column for character in self.characters])
        self.input_height = max([character.input_coord.row for character in self.characters])
        self.output_area = OutputArea(min(self.height - 1, self.input_height), self.input_width)
        self.characters = [
            character for character in self.characters if character.input_coord.row <= self.output_area.top
        ]
        self.animation_rate = args.animation_rate
        self.last_time_printed = time.time()
        self._update_terminal_state()

        self._prep_outputarea()

    def _get_terminal_dimensions(self) -> tuple[int, int]:
        """Gets the terminal dimensions.

        Returns:
            tuple[int, int]: terminal width and height
        """
        try:
            terminal_width, terminal_height = shutil.get_terminal_size()
        except OSError:
            # If the terminal size cannot be determined, return default values
            return 80, 24
        return terminal_width, terminal_height

    @staticmethod
    def get_piped_input() -> str:
        """Gets the piped input from stdin.

        Returns:
            str: string from stdin
        """
        if not sys.stdin.isatty():
            input_data = sys.stdin.read()
            return input_data
        else:
            return ""

    def _decompose_input(self, use_xterm_colors: bool, no_color: bool) -> list[EffectCharacter]:
        """Decomposes the output into a list of Character objects containing the symbol and its row/column coordinates
        relative to the input display location.

        Coordinates are relative to the cursor row position at the time of execution. 1,1 is the bottom left corner of the row
        above the cursor.

        Args:
            use_xterm_colors (bool): whether to convert colors to the closest XTerm-256 color

        Returns:
            list[Character]: list of EffectCharacter objects
        """
        wrapped_lines = []
        if not self.input_data.strip():
            self.input_data = "No Input."
        input_lines = self.input_data.splitlines()
        for line in input_lines:
            while len(line) > self.width:
                wrapped_lines.append(line[: self.width])
                line = line[self.width :]
            wrapped_lines.append(line)
        input_height = len(wrapped_lines)
        input_characters = []
        for row, line in enumerate(wrapped_lines):
            for column, symbol in enumerate(line):
                if symbol != " ":
                    character = EffectCharacter(symbol, column + 1, input_height - row, self)
                    character.animator.use_xterm_colors = use_xterm_colors
                    character.animator.no_color = no_color
                    input_characters.append(character)
        return input_characters

    def _update_terminal_state(self):
        """Update the internal representation of the terminal state with the current position
        of all active characters.
        """
        rows = [[" " for _ in range(self.output_area.right)] for _ in range(self.output_area.top)]
        for character in self.characters:
            if character.is_active:
                try:
                    # do not allow characters to wrap around the terminal via negative coordinates
                    if character.motion.current_coord.row <= 0 or character.motion.current_coord.column <= 0:
                        continue
                    rows[character.motion.current_coord.row - 1][
                        character.motion.current_coord.column - 1
                    ] = character.symbol
                except IndexError:
                    # ignore characters that are outside the output area
                    pass
        terminal_state = ["".join(row) for row in rows]
        self.terminal_state = terminal_state

    def _prep_outputarea(self) -> None:
        """Prepares the terminal for the effect by adding empty lines above."""
        print("\n" * self.output_area.top)

    def print(self):
        """Prints the current terminal state to stdout while preserving the cursor position."""
        self._update_terminal_state()
        time_since_last_print = time.time() - self.last_time_printed
        if time_since_last_print < self.animation_rate:
            time.sleep(self.animation_rate - time_since_last_print)
        output = "\n".join(self.terminal_state[::-1])
        sys.stdout.write(ansitools.DEC_SAVE_CURSOR_POSITION())
        sys.stdout.write(ansitools.MOVE_CURSOR_UP(self.output_area.top))
        sys.stdout.write(ansitools.MOVE_CURSOR_TO_COLUMN(1))
        sys.stdout.write(output)
        sys.stdout.write(ansitools.DEC_RESTORE_CURSOR_POSITION())
        sys.stdout.flush()
        self.last_time_printed = time.time()



import argparse

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects import base_character, base_effect
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, argtypes


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "effect_name",
        formatter_class=argtypes.CustomFormatter,
        help="effect_description",
        description="effect_description",
        epilog=f"""{argtypes.EASING_EPILOG}

Example: effect_example""",
    )
    effect_parser.set_defaults(effect_class=NamedEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time, in seconds, between animation steps.",
    )
    effect_parser.add_argument(
        "--color-single",  # make more descriptive
        type=argtypes.valid_color,
        default=0,
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the ___.",
    )
    effect_parser.add_argument(
        "--color-list",  # make more descriptive
        type=argtypes.valid_color,
        nargs="*",
        default=0,
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Up to N colors for the __.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the final character.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=1,
        metavar="(float > 0)",
        help="Speed of the ___. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="IN_OUT_SINE",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement.",
    )


class NamedEffect(base_effect.Effect):
    """Effect that ___."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        super().__init__(terminal, args)

    def prepare_data(self) -> None:
        """Prepares the data for the effect by ___."""

        for character in self.terminal.characters:
            pass
            # do something with the data if needed (sort, adjust positions, etc)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        while self.pending_chars or self.animating_chars:
            self.terminal.print()
            self.animate_chars()

            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animator.is_active_scene_complete() or animating_char.motion.movement_complete()
            ]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation."""
        for animating_char in self.animating_chars:
            animating_char.animator.step_animation()
            animating_char.motion.move()
            if animating_char.motion.movement_complete():
                animating_char.symbol = animating_char.input_symbol



import argparse
from enum import Enum, auto

from terminaltexteffects import base_character, base_effect
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import argtypes


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "columnslide",
        formatter_class=argtypes.CustomFormatter,
        help="Slides each column into place from the outside to the middle.",
        description="columnslide | Slides each column into place from the outside to the middle.",
        epilog=f"""{argtypes.EASING_EPILOG}
            
Example: terminaltexteffects columnslide -a 0.003 --slide-direction up --easing IN_OUT_SINE --movement-speed 0.2 --column-gap 5""",
    )
    effect_parser.set_defaults(effect_class=ColumnSlide)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.003,
        help="Time to sleep, in seconds, between animation steps.",
    )
    effect_parser.add_argument(
        "--column-gap",
        default=5,
        type=argtypes.valid_gap,
        help="Number of characters to wait before adding a new column.",
    )
    effect_parser.add_argument(
        "--slide-direction",
        default="down",
        choices=["up", "down"],
        help="Direction the text will slide.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.2,
        metavar="(float > 0)",
        help="Character movement speed. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="IN_OUT_SINE",
        type=argtypes.valid_ease,
        help="Easing function to use for column movement.",
    )


class SlideDirection(Enum):
    UP = auto()
    DOWN = auto()


class ColumnSlide(base_effect.Effect):
    """Effect that slides each column into place."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        """Effect that slides each column into place.

        Args:
            input_data (str): string from stdin
            args (argparse.Namespace): arguments from argparse
        """
        super().__init__(terminal, args)
        self.column_gap: int = args.column_gap
        if args.slide_direction == "down":
            self.slide_direction = SlideDirection.DOWN
        else:
            self.slide_direction = SlideDirection.UP

    def prepare_data(self) -> None:
        """Prepares the data for the effect by grouping the characters by column and setting the starting
        coordinate."""

        self.columns = self.input_by_column()
        if self.slide_direction == SlideDirection.DOWN:
            for column_list in self.columns.values():
                column_list.reverse()
        for column in self.columns.values():
            for character in column:
                character.is_active = False
                if self.slide_direction == SlideDirection.DOWN:
                    character.motion.set_coordinate(character.input_coord.column, self.terminal.output_area.top)
                else:
                    character.motion.set_coordinate(character.input_coord.column, self.terminal.output_area.bottom)
                character.motion.new_waypoint(
                    "input_coord",
                    character.input_coord.column,
                    character.input_coord.row,
                    speed=self.args.movement_speed,
                    ease=self.args.easing,
                )
                character.motion.activate_waypoint("input_coord")

    def get_next_column(self) -> list[base_character.EffectCharacter]:
        """Gets the next column of characters to animate.

        Returns:
            list[character.EffectCharacter]: The next column of characters to animate.
        """
        if len(self.columns) % 2:
            next_column = self.columns.pop(max(self.columns.keys()))
        else:
            next_column = self.columns.pop(min(self.columns.keys()))
        return next_column

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        active_columns: list[list[base_character.EffectCharacter]] = []
        active_columns.append(self.get_next_column())
        column_delay_countdown = self.column_gap
        self.terminal.print()
        while active_columns or self.animating_chars or self.columns:
            if (column_delay_countdown == 0 and self.columns) or (self.columns and not active_columns):
                active_columns.append(self.get_next_column())
                column_delay_countdown = self.column_gap
            else:
                if self.columns:
                    column_delay_countdown -= 1
            for column in active_columns:
                if column:
                    next_character = column.pop(0)
                    next_character.is_active = True
                    self.animating_chars.append(next_character)
            self.animate_chars()

            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.motion.movement_complete()
            ]
            active_columns = [column for column in active_columns if column]
            self.terminal.print()

    def animate_chars(self) -> None:
        """Animates the characters by calling the tween method and printing the characters to the terminal."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()



import random, argparse
import terminaltexteffects.utils.terminal as terminal
from terminaltexteffects.base_character import EffectCharacter


class Effect:
    """Generic class for all effects. Derive from this class to create a new effect."""

    def __init__(self, terminal: terminal.Terminal, args: argparse.Namespace):
        """Initializes the Effect class.

        Args:
            terminal (terminal.Terminal): a Terminal object
        """
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []

    def random_column(self) -> int:
        """Returns a random column position."""
        return random.randint(1, self.terminal.output_area.right)

    def random_row(self) -> int:
        """Returns a random row position."""
        return random.randint(1, self.terminal.output_area.top)

    def input_by_row(self) -> dict[int, list[EffectCharacter]]:
        """Returns a dict of rows of EffectCharacters where the key is the row index.
        Returns:
            dict[int,list[EffectCharacter]]: dict of rows of EffectCharacters
        """
        rows: dict[int, list[EffectCharacter]] = dict()
        for row_index in range(self.terminal.output_area.top + 1):
            characters_in_row = [
                character for character in self.terminal.characters if character.input_coord.row == row_index
            ]
            if characters_in_row:
                rows[row_index] = characters_in_row
        return rows

    def input_by_column(self) -> dict[int, list[EffectCharacter]]:
        """Returns a dict columns of EffectCharacters where the key is the column index.

        Returns:
            dict[int,list[EffectCharacter]]: dict of columns of EffectCharacters
        """
        columns: dict[int, list[EffectCharacter]] = dict()
        for column_index in range(self.terminal.output_area.right + 1):
            characters_in_column = [
                character for character in self.terminal.characters if character.input_coord.column == column_index
            ]
            if characters_in_column:
                columns[column_index] = characters_in_column
        return columns



import typing
from dataclasses import dataclass, field

from terminaltexteffects.utils import ansitools, colorterm, hexterm

if typing.TYPE_CHECKING:
    from terminaltexteffects import base_character


@dataclass
class GraphicalEffect:
    """A class for storing terminal graphical modes and a color.

    Supported graphical modes:
    bold, dim, italic, underline, blink, inverse, hidden, strike

    Args:
        symbol (str): the symbol to show
        bold (bool): bold mode
        dim (bool): dim mode
        italic (bool): italic mode
        underline (bool): underline mode
        blink (bool): blink mode
        reverse (bool): reverse mode
        hidden (bool): hidden mode
        strike (bool): strike mode
        color (int | str): color code
    """

    symbol: str
    bold: bool = False
    dim: bool = False
    italic: bool = False
    underline: bool = False
    blink: bool = False
    reverse: bool = False
    hidden: bool = False
    strike: bool = False
    color: int | str | None = None

    def __post_init__(self):
        self.format_symbol()

    def disable_modes(self) -> None:
        """Disables all graphical modes."""
        self.bold = False
        self.dim = False
        self.italic = False
        self.underline = False
        self.blink = False
        self.reverse = False
        self.hidden = False
        self.strike = False

    def format_symbol(self) -> None:
        """Formats the symbol for printing."""
        formatting_string = ""
        if self.bold:
            formatting_string += ansitools.APPLY_BOLD()
        if self.italic:
            formatting_string += ansitools.APPLY_ITALIC()
        if self.underline:
            formatting_string += ansitools.APPLY_UNDERLINE()
        if self.blink:
            formatting_string += ansitools.APPLY_BLINK()
        if self.reverse:
            formatting_string += ansitools.APPLY_REVERSE()
        if self.hidden:
            formatting_string += ansitools.APPLY_HIDDEN()
        if self.strike:
            formatting_string += ansitools.APPLY_STRIKETHROUGH()
        if self.color is not None:
            formatting_string += colorterm.fg(self.color)

        self.symbol = f"{formatting_string}{self.symbol}{ansitools.RESET_ALL() if formatting_string else ''}"


@dataclass
class Sequence:
    """An Sequence is a graphicaleffect with a duration.
    Args:
        graphical_effect (GraphicalEffect): a GraphicalEffect object containing the graphical modes and color of the character
        duration (int): the number of animation steps to use the Sequence
    """

    graphical_effect: GraphicalEffect
    duration: int

    def __post_init__(self):
        self.frames_played = 0
        self.symbol = self.graphical_effect.symbol


@dataclass
class Scene:
    """A Scene is a list of Sequences.

    Args:
        id (str): the id of the Scene
        sequences (list[Sequence]): a list of Sequence objects
        is_looping (bool): whether the Scene should loop
    """

    id: str
    sequences: list[Sequence] = field(default_factory=list)
    is_looping: bool = False

    def __post_init__(self):
        self.played_sequences = []

    def add_sequence(self, graphicaleffect: GraphicalEffect, duration: int) -> None:
        """Adds a Sequence to the Scene.

        Args:
            graphicaleffect (GraphicalEffect): a GraphicalEffect object containing the graphical modes and color of the character
            duration (int): the number of animation steps to use the Sequence
        """
        self.sequences.append(Sequence(graphicaleffect, duration))

    def get_next_symbol(self) -> str:
        """Returns the next symbol in the Scene.

        Returns:
            str: the next symbol in the Scene
        """
        current_sequence = self.sequences[0]
        next_symbol = current_sequence.symbol
        current_sequence.frames_played += 1
        if current_sequence.frames_played == current_sequence.duration:
            current_sequence.frames_played = 0
            self.played_sequences.append(self.sequences.pop(0))
            if self.is_looping and not self.sequences:
                self.sequences.extend(self.played_sequences)
                self.played_sequences.clear()
        return next_symbol

    def reset_scene(self) -> None:
        """Resets the Scene."""
        for sequence in self.sequences:
            sequence.frames_played = 0
            self.played_sequences.append(sequence)
        self.sequences.clear()
        self.sequences.extend(self.played_sequences)
        self.played_sequences.clear()


class Animator:
    def __init__(self, character: "base_character.EffectCharacter"):
        """Animator handles the animations of a character. It contains a list of Scenes, and the active scene id. GraphicalEffects are
        added to the Scenes, and the Animator returns the next symbol in the active scene.

        Args:
            character (base_character.EffectCharacter): the EffectCharacter object to animate
        """
        self.character = character
        self.scenes: dict[str, Scene] = {}
        self.active_scene: Scene = None
        self.use_xterm_colors: bool = False
        self.no_color: bool = False
        self.xterm_color_map: dict[str, int] = {}

    def add_effect_to_scene(
        self, scene_id: str, symbol: str | None = None, color: int | str | None = None, duration: int = 1
    ) -> None:
        """Add a graphical effect to a scene. If the scene doesn't exist, it will be created.

        Args:
            scene_id (str): ID of the scene to which the effect will be added
            symbol (str): Symbol to display, if None, the character's input symbol will be used
            color (int | str): XTerm color code (0-255) or hex color code (e.g. ffffff)
            duration (int): Number of animation steps to display the effect, defaults to 1. Minimum 1.
        """
        if duration < 1:
            raise ValueError("duration must be greater than 0")
        if scene_id not in self.scenes:
            new_scene = Scene(scene_id)
            self.scenes[scene_id] = new_scene
        if not symbol:
            symbol = self.character.input_symbol
        if not self.no_color:
            if self.use_xterm_colors and isinstance(color, str):
                if color in self.xterm_color_map:
                    color = self.xterm_color_map[color]
                else:
                    xterm_color = hexterm.hex_to_xterm(color)
                    self.xterm_color_map[color] = xterm_color
                    color = xterm_color
        else:
            color = None
        graphicaleffect = GraphicalEffect(symbol, color=color)
        self.scenes[scene_id].add_sequence(graphicaleffect, duration)

    def is_active_scene_complete(self) -> bool:
        """Returns whether the active scene is complete. A scene is complete if all sequences have been played.
        Looping scenes are always complete.

        Returns:
            bool: True if complete, False otherwise
        """
        if not self.active_scene:
            return True
        if not self.active_scene.sequences or self.active_scene.is_looping:
            return True
        else:
            return False

    def step_animation(self) -> None:
        """Apply the next symbol in the scene to the character. If a scene order exists, the next scene
        will be activated when the current scene is complete."""
        if self.active_scene and self.active_scene.sequences:
            self.character.symbol = self.active_scene.get_next_symbol()
            if self.is_active_scene_complete():
                self.character.event_handler.handle_event(
                    self.character.event_handler.Event.SCENE_COMPLETE, self.active_scene.id
                )

    def reset_scene(self, scene_id: str) -> None:
        """Resets the Scene.

        Args:
            scene_name (str): the ID of the Scene to reset
        """
        if scene_id in self.scenes:
            self.scenes[scene_id].reset_scene()

    def activate_scene(self, scene_id: str) -> None:
        """Sets the active scene.

        Args:
            scene_id (str): the ID of the Scene to set as active
        """
        if scene_id in self.scenes:
            self.active_scene = self.scenes[scene_id]
            self.step_animation()
        else:
            raise ValueError(f"Scene {scene_id} does not exist")


def gradient(start_color: str | int, end_color: str | int, steps: int) -> list[str]:
    """Calculate a gradient of colors between two colors using linear interpolation.

    Args:
        start_color (str|int): RGB hex color string or XTerm-256 color code
        end_color (str|int): RGB hex color string or XTerm-256 color code
        steps (int): Number of intermediate colors to calculate

    Returns:
        list[str]: List (length=steps + 2) of RGB hex color strings
    """
    if isinstance(start_color, int):
        start_color = hexterm.xterm_to_hex(start_color)
    if isinstance(end_color, int):
        end_color = hexterm.xterm_to_hex(end_color)
    start_color_ints = colorterm._hex_to_int(start_color)
    end_color_ints = colorterm._hex_to_int(end_color)
    gradient_colors = []
    gradient_colors.append(start_color)
    red_delta = (end_color_ints[0] - start_color_ints[0]) // steps
    green_delta = (end_color_ints[1] - start_color_ints[1]) // steps
    blue_delta = (end_color_ints[2] - start_color_ints[2]) // steps
    for i in range(steps):
        red = start_color_ints[0] + (red_delta * i)
        green = start_color_ints[1] + (green_delta * i)
        blue = start_color_ints[2] + (blue_delta * i)

        red = max(0, min(red, 255))
        green = max(0, min(green, 255))
        blue = max(0, min(blue, 255))

        gradient_colors.append(f"{red:02x}{green:02x}{blue:02x}")
    gradient_colors.append(end_color)
    return gradient_colors



import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
import terminaltexteffects.utils.graphics as graphics
from terminaltexteffects import base_character, base_effect
from terminaltexteffects.utils.terminal import Terminal


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "shootingstar",
        formatter_class=argtypes.CustomFormatter,
        help="Displays the text as a falling star toward the final coordinate of the character.",
        description="shootingstar | Displays the text as a falling star toward the final coordinate of the character.",
        epilog=f"""{argtypes.EASING_EPILOG}
        
Example: terminaltexteffects shootingstar -a 0.01""",
    )
    effect_parser.set_defaults(effect_class=ShootingStarEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time between animation steps. ",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=1,
        metavar="(float > 0)",
        help="Movement speed of the characters. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="OUT_CIRC",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement.",
    )


class ShootingStarEffect(base_effect.Effect):
    """Effect that display the text as a falling star toward the final coordinate of the character."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        super().__init__(terminal, args)
        self.group_by_row: dict[int, list[base_character.EffectCharacter | None]] = {}

    def prepare_data(self) -> None:
        """Prepares the data for the effect by sorting by row and assigning the star symbol."""

        for character in self.terminal.characters:
            character.is_active = False
            character.animator.add_effect_to_scene("star", "*", random.randint(1, 10), 1)
            character.animator.add_effect_to_scene("final", character.input_symbol, duration=1)
            character.motion.set_coordinate(self.random_column(), self.terminal.output_area.top)
            character.motion.new_waypoint(
                "input_coord",
                character.input_coord.column,
                character.input_coord.row,
                speed=self.args.movement_speed,
                ease=self.args.easing,
            )
            character.motion.activate_waypoint("input_coord")
            character.animator.activate_scene("star")
            character.event_handler.register_event(
                character.event_handler.Event.WAYPOINT_REACHED,
                "input_coord",
                character.event_handler.Action.ACTIVATE_SCENE,
                "final",
            )
            self.pending_chars.append(character)
        for character in sorted(self.pending_chars, key=lambda c: c.input_coord.row):
            if character.input_coord.row not in self.group_by_row:
                self.group_by_row[character.input_coord.row] = []
            self.group_by_row[character.input_coord.row].append(character)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        self.pending_chars.clear()
        while self.group_by_row or self.animating_chars or self.pending_chars:
            if not self.pending_chars and self.group_by_row:
                self.pending_chars.extend(self.group_by_row.pop(min(self.group_by_row.keys())))  # type: ignore
            if self.pending_chars:
                for _ in range(random.randint(1, 3)):
                    if self.pending_chars:
                        next_character = self.pending_chars.pop(random.randint(0, len(self.pending_chars) - 1))
                        next_character.is_active = True
                        self.animating_chars.append(next_character)
                    else:
                        break
            self.animate_chars()
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animator.is_active_scene_complete()
                or not animating_char.motion.movement_complete()
            ]
            self.terminal.print()

    def animate_chars(self) -> None:
        """Animates the characters by calling the tween method and printing the characters to the terminal."""
        for animating_char in self.animating_chars:
            animating_char.animator.step_animation()
            animating_char.motion.move()



import typing, math
from dataclasses import dataclass
from enum import Enum, auto
import terminaltexteffects.utils.easing as easing

if typing.TYPE_CHECKING:
    from terminaltexteffects import base_character


@dataclass
class Coord:
    """A coordinate with row and column values.

    Args:
        column (int): column value
        row (int): row value"""

    column: int
    row: int


@dataclass
class Waypoint:
    """A coordinate, speed, and easing function.

    Args:
        waypoint_id (str): unique identifier for the waypoint
        coord (Coord): coordinate
        speed (float): character speed at the waypoint
        ease (Ease): easing function for character movement. Defaults to None.
        next_waypoint (Waypoint): next waypoint to move to. Defaults to None.
    """

    waypoint_id: str
    coord: Coord
    speed: float = 1.0
    ease: easing.Ease | None = None


class Motion:
    def __init__(self, character: "base_character.EffectCharacter"):
        """Motion class for managing the movement of a character. Waypoints store target coordinates, speeds, and easing functions. The character
        is moved from one waypoint to the next until all waypoints have been reached. If no waypoints remain and character is not at the
        input coordinate, a new waypoint for the input coodinate is added and the character is moved to the new waypoint.

        Args:
            character (base_character.EffectCharacter): The character to move.
        """
        self.character = character
        self.current_coord: Coord = Coord(character.input_coord.column, character.input_coord.row)
        self.previous_coord: Coord = Coord(-1, -1)
        self.speed: float = 1.0
        self.waypoints: dict[str, Waypoint] = {}
        self.active_waypoint: Waypoint | None = None
        self.origin_waypoint: Waypoint | None = None
        self.inter_waypoint_distance: float = 0
        self.inter_waypoint_max_steps: int = 0
        self.inter_waypoint_current_step: int = 0

    def _distance(self, column1: int, row1: int, column2: int, row2: int) -> float:
        """Returns the distance between two coordinates.

        Args:
            column1 (int): column of the first waypoint
            row1 (int): row of the first waypoint
            column2 (int): column of the second waypoint
            row2 (int): row of the second waypoint

        Returns:
            float: distance between the two waypoints
        """
        column_diff = column2 - column1
        row_diff = row2 - row1
        return math.hypot(column_diff, row_diff)

    def _point_at_distance(self, distance: float) -> Coord:
        """Returns the coordinate at the given distance along the line defined by the origin waypoint and current waypoint.

        Args:
            distance (float): distance

        Returns:
            Coord: Coordinate at the given distance.
        """
        if not distance or not self.origin_waypoint or not self.active_waypoint:
            return self.current_coord
        t = distance / self.inter_waypoint_distance
        next_column, next_row = (
            ((1 - t) * self.origin_waypoint.coord.column + t * self.active_waypoint.coord.column),
            ((1 - t) * self.origin_waypoint.coord.row + t * self.active_waypoint.coord.row),
        )
        return Coord(round(next_column), round(next_row))

    def find_points_on_circle(self, origin: tuple[int, int], radius: int, num_points: int) -> Coord:
        """Finds points on a circle.

        Args:
            origin (tuple[int, int]): origin of the circle
            radius (int): radius of the circle
            num_points (int): number of points to find

        Returns:
            list (Coord): list of Coord points on the circle
        """
        points = []
        for i in range(num_points):
            angle = 2 * math.pi * i / num_points
            x = origin[0] + radius * math.cos(angle)
            y = origin[1] + radius * math.sin(angle)
            points.append(Coord(round(x), round(y)))
        return points

    def set_coordinate(self, column: int, row: int) -> None:
        """Sets the current coordinate to the given coordinate.

        Args:
            column (int): column
            row (int): row
        """
        self.current_coord = Coord(column, row)

    def new_waypoint(
        self, waypoint_id: str, column: int, row: int, speed: float = 1, ease: easing.Ease | None = None
    ) -> None:
        """Appends a new waypoint to the waypoints dictionary.

        Args:
            waypoint_id (str): unique identifier for the waypoint
            column (int): column
            row (int): row
            speed (float): speed
            ease (Ease| None): easing function for character movement. Defaults to None.
        """
        self.waypoints[waypoint_id] = Waypoint(waypoint_id, Coord(column, row), speed, ease)

    def movement_complete(self) -> bool:
        """Returns whether the character has reached the final coordinate.

        Returns:
            bool: True if the character has reached the final coordinate and has no active waypoint, False otherwise.
        """
        if (
            self.inter_waypoint_current_step == self.inter_waypoint_max_steps
            and self.current_coord == self.character.input_coord
        ):
            return True
        return False

    def _ease_movement(self, easing_func: easing.Ease) -> float:
        """Returns the percentage of total distance that should be moved based on the easing function.

        Args:
            easing_func (easing.Ease): The easing function to use.

        Returns:
            float: The percentage of total distance to move.
        """
        easing_function_map = {
            easing.Ease.IN_SINE: easing.in_sine,
            easing.Ease.OUT_SINE: easing.out_sine,
            easing.Ease.IN_OUT_SINE: easing.in_out_sine,
            easing.Ease.IN_QUAD: easing.in_quad,
            easing.Ease.OUT_QUAD: easing.out_quad,
            easing.Ease.IN_OUT_QUAD: easing.in_out_quad,
            easing.Ease.IN_CUBIC: easing.in_cubic,
            easing.Ease.OUT_CUBIC: easing.out_cubic,
            easing.Ease.IN_OUT_CUBIC: easing.in_out_cubic,
            easing.Ease.IN_QUART: easing.in_quart,
            easing.Ease.OUT_QUART: easing.out_quart,
            easing.Ease.IN_OUT_QUART: easing.in_out_quart,
            easing.Ease.IN_QUINT: easing.in_quint,
            easing.Ease.OUT_QUINT: easing.out_quint,
            easing.Ease.IN_OUT_QUINT: easing.in_out_quint,
            easing.Ease.IN_EXPO: easing.in_expo,
            easing.Ease.OUT_EXPO: easing.out_expo,
            easing.Ease.IN_OUT_EXPO: easing.in_out_expo,
            easing.Ease.IN_CIRC: easing.in_circ,
            easing.Ease.OUT_CIRC: easing.out_circ,
            easing.Ease.IN_OUT_CIRC: easing.in_out_circ,
            easing.Ease.IN_BACK: easing.in_back,
            easing.Ease.OUT_BACK: easing.out_back,
            easing.Ease.IN_OUT_BACK: easing.in_out_back,
            easing.Ease.IN_ELASTIC: easing.in_elastic,
            easing.Ease.OUT_ELASTIC: easing.out_elastic,
            easing.Ease.IN_OUT_ELASTIC: easing.in_out_elastic,
            easing.Ease.IN_BOUNCE: easing.in_bounce,
            easing.Ease.OUT_BOUNCE: easing.out_bounce,
            easing.Ease.IN_OUT_BOUNCE: easing.in_out_bounce,
        }
        elapsed_step_ratio = self.inter_waypoint_current_step / self.inter_waypoint_max_steps
        return easing_function_map[easing_func](elapsed_step_ratio)

    def activate_waypoint(self, waypoint_id: str) -> None:
        """Sets the current waypoint to the waypoint with the given waypoint_id and sets
        the speed, distance, and step values for the new waypoint.

        Args:
            waypoint_id (str): unique identifier for the waypoint
        """
        if not self.active_waypoint:
            self.origin_waypoint = Waypoint("-1", Coord(self.current_coord.column, self.current_coord.row))
        else:
            self.origin_waypoint = self.active_waypoint
        self.active_waypoint = self.waypoints[waypoint_id]
        self.speed = self.active_waypoint.speed
        self.inter_waypoint_distance = self._distance(
            self.current_coord.column,
            self.current_coord.row,
            self.active_waypoint.coord.column,
            self.active_waypoint.coord.row,
        )
        self.inter_waypoint_current_step = 0
        if self.speed > self.inter_waypoint_distance:
            self.speed = max(self.inter_waypoint_distance, 1)
        self.inter_waypoint_max_steps = round(self.inter_waypoint_distance / self.speed)

    def deactivate_waypoint(self) -> None:
        """Unsets the current waypoint."""
        self.active_waypoint = None
        self.inter_waypoint_distance = 0
        self.inter_waypoint_max_steps = 0
        self.inter_waypoint_current_step = 0

    def move(self) -> None:
        """Moves the character one step closer to the target position based on an easing function if present, otherwise linearly."""
        # preserve previous coordinate to allow for clearing the location in the terminal
        self.previous_coord.column, self.previous_coord.row = (
            self.current_coord.column,
            self.current_coord.row,
        )
        if not self.active_waypoint:
            return
        if self.inter_waypoint_current_step < self.inter_waypoint_max_steps:
            self.inter_waypoint_current_step += 1

        if self.inter_waypoint_distance:
            if self.active_waypoint and self.active_waypoint.ease:
                easing_factor = self._ease_movement(self.active_waypoint.ease)
                distance_to_move = easing_factor * self.inter_waypoint_distance
            else:
                linear_factor = self.inter_waypoint_current_step / self.inter_waypoint_max_steps
                distance_to_move = linear_factor * self.inter_waypoint_distance
            self.current_coord = self._point_at_distance(distance_to_move)

        if self.inter_waypoint_current_step == self.inter_waypoint_max_steps:
            self.character.event_handler.handle_event(
                self.character.event_handler.Event.WAYPOINT_REACHED, self.active_waypoint.waypoint_id
            )



"""Effect that draws the characters spawning at varying rates from a single point."""

import argparse
import random
from enum import Enum, auto

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects import base_character, base_effect
from terminaltexteffects.utils import graphics, motion
from terminaltexteffects.utils.terminal import Terminal


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "spray",
        help="Draws the characters spawning at varying rates from a single point.",
        description="spray | Draws the characters spawning at varying rates from a single point.",
        epilog=f"""{argtypes.EASING_EPILOG}
        
Example: terminaltexteffects spray -a 0.01 --spray-position center""",
    )
    effect_parser.set_defaults(effect_class=SprayEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time, in seconds, between animation steps.",
    )
    effect_parser.add_argument(
        "--spray-colors",
        type=argtypes.valid_color,
        nargs="*",
        default=0,
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="List of colors for the character spray. Colors are randomly chosen from the list.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the final character.",
    )
    effect_parser.add_argument(
        "--spray-position",
        default="e",
        choices=["n", "ne", "e", "se", "s", "sw", "w", "nw", "center"],
        help="Position for the spray origin.",
    )
    effect_parser.add_argument(
        "--spray-volume",
        default=5,
        type=argtypes.positive_int,
        metavar="(int > 0)",
        help="Maximum number of characters to spray at a time.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.7,
        metavar="(float > 0)",
        help="Movement speed of the characters. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="OUT_EXPO",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement.",
    )


class SprayPosition(Enum):
    """Position for the spray origin."""

    N = auto()
    NE = auto()
    E = auto()
    SE = auto()
    S = auto()
    SW = auto()
    W = auto()
    NW = auto()
    CENTER = auto()


class SprayEffect(base_effect.Effect):
    """Effect that draws the characters spawning at varying rates from a single point."""

    def __init__(
        self,
        terminal: Terminal,
        args: argparse.Namespace,
    ):
        """Effect that draws the characters spawning at varying rates from a single point.

        Args:
            terminal (Terminal): terminal to use for the effect
            args (argparse.Namespace): arguments from argparse
        """
        super().__init__(terminal, args)
        self.spray_position = {
            "n": SprayPosition.N,
            "ne": SprayPosition.NE,
            "e": SprayPosition.E,
            "se": SprayPosition.SE,
            "s": SprayPosition.S,
            "sw": SprayPosition.SW,
            "w": SprayPosition.W,
            "nw": SprayPosition.NW,
            "center": SprayPosition.CENTER,
        }.get(args.spray_position, SprayPosition.E)
        self.spray_colors = args.spray_colors
        self.final_color = args.final_color

    def prepare_data(self) -> None:
        """Prepares the data for the effect by starting all of the characters from a point based on SparklerPosition."""
        spray_origin_map = {
            SprayPosition.CENTER: (
                self.terminal.output_area.right // 2,
                self.terminal.output_area.top // 2,
            ),
            SprayPosition.N: (self.terminal.output_area.right // 2, self.terminal.output_area.top),
            SprayPosition.NW: (self.terminal.output_area.left, self.terminal.output_area.top),
            SprayPosition.W: (self.terminal.output_area.left, self.terminal.output_area.top // 2),
            SprayPosition.SW: (self.terminal.output_area.left, self.terminal.output_area.bottom),
            SprayPosition.S: (self.terminal.output_area.right // 2, self.terminal.output_area.bottom),
            SprayPosition.SE: (self.terminal.output_area.right - 1, self.terminal.output_area.bottom),
            SprayPosition.E: (self.terminal.output_area.right - 1, self.terminal.output_area.top // 2),
            SprayPosition.NE: (self.terminal.output_area.right - 1, self.terminal.output_area.top),
        }

        for character in self.terminal.characters:
            character.motion.set_coordinate(*spray_origin_map[self.spray_position])
            character.motion.new_waypoint(
                "input_coord",
                character.input_coord.column,
                character.input_coord.row,
                speed=self.args.movement_speed,
                ease=self.args.easing,
            )
            if self.spray_colors:
                spray_color = random.choice(self.spray_colors)
                spray_gradient = graphics.gradient(spray_color, self.final_color, 7)
                for color in spray_gradient:
                    character.animator.add_effect_to_scene("droplet", character.input_symbol, color, 40)
                character.animator.activate_scene("droplet")
            character.motion.activate_waypoint("input_coord")
            self.pending_chars.append(character)
        random.shuffle(self.pending_chars)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        while self.pending_chars or self.animating_chars:
            if self.pending_chars:
                for _ in range(random.randint(1, self.args.spray_volume)):
                    if self.pending_chars:
                        next_character = self.pending_chars.pop()
                        next_character.is_active = True
                        self.animating_chars.append(next_character)

            self.animate_chars()
            self.terminal.print()
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animator.is_active_scene_complete()
                or not animating_char.motion.movement_complete()
            ]

    def animate_chars(self) -> None:
        for animating_char in self.animating_chars:
            animating_char.animator.step_animation()
            animating_char.motion.move()



"""Effect that pours the characters into position from the top, bottom, left, or right."""

import argparse
from enum import Enum, auto

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects import base_effect
from terminaltexteffects.utils.terminal import Terminal


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "pour",
        formatter_class=argtypes.CustomFormatter,
        help="Pours the characters into position from the given direction.",
        description="pour | Pours the characters into position from the given direction.",
        epilog=f"""{argtypes.EASING_EPILOG}
        
Example: terminaltexteffects pour -a 0.004 --pour-direction down""",
    )
    effect_parser.set_defaults(effect_class=PourEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.004,
        help="Time, in seconds, between animation steps.",
    )
    effect_parser.add_argument(
        "--pour-direction",
        default="down",
        choices=["up", "down", "left", "right"],
        help="Direction the text will pour.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.2,
        metavar="(float > 0)",
        help="Movement speed of the characters. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="IN_QUAD",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement.",
    )


class PourDirection(Enum):
    UP = auto()
    DOWN = auto()
    LEFT = auto()
    RIGHT = auto()


class PourEffect(base_effect.Effect):
    """Effect that pours the characters into position from the top, bottom, left, or right."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        super().__init__(terminal, args)
        self.pour_direction = {
            "down": PourDirection.DOWN,
            "up": PourDirection.UP,
            "left": PourDirection.LEFT,
            "right": PourDirection.RIGHT,
        }.get(args.pour_direction, PourDirection.DOWN)

    def prepare_data(self) -> None:
        """Prepares the data for the effect by sorting the characters by the pour direction."""
        sort_map = {
            PourDirection.DOWN: lambda character: character.input_coord.row,
            PourDirection.UP: lambda character: -character.input_coord.row,
            PourDirection.LEFT: lambda character: character.input_coord.column,
            PourDirection.RIGHT: lambda character: -character.input_coord.column,
        }
        self.terminal.characters.sort(key=sort_map[self.pour_direction])
        for character in self.terminal.characters:
            character.is_active = False
            if self.pour_direction == PourDirection.DOWN:
                character.motion.set_coordinate(character.input_coord.column, self.terminal.output_area.top)
            elif self.pour_direction == PourDirection.UP:
                character.motion.set_coordinate(character.input_coord.column, self.terminal.output_area.bottom)
            elif self.pour_direction == PourDirection.LEFT:
                character.motion.set_coordinate(self.terminal.output_area.right, character.input_coord.row)
            elif self.pour_direction == PourDirection.RIGHT:
                character.motion.set_coordinate(self.terminal.output_area.left, character.input_coord.row)
            character.motion.new_waypoint(
                "input_coord",
                character.input_coord.column,
                character.input_coord.row,
                speed=self.args.movement_speed,
                ease=self.args.easing,
            )
            character.motion.activate_waypoint("input_coord")
            self.pending_chars.append(character)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        self.terminal.print()
        while self.pending_chars or self.animating_chars:
            if self.pending_chars:
                next_character = self.pending_chars.pop(0)
                next_character.is_active = True
                self.animating_chars.append(next_character)
            self.animate_chars()
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.motion.movement_complete()
            ]
            self.terminal.print()

    def animate_chars(self) -> None:
        """Animates the sliding characters."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()



import argparse
from enum import Enum, auto

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects import base_character, base_effect
from terminaltexteffects.utils.terminal import Terminal


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "rowslide",
        formatter_class=argtypes.CustomFormatter,
        help="Slides each row into place.",
        description="rowslide | Slides each row into place.",
        epilog=f"""{argtypes.EASING_EPILOG}
        
Example: terminaltexteffects rowslide -a 0.003 --slide-direction left""",
    )
    effect_parser.set_defaults(effect_class=RowSlide)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time, in seconds, between animation steps.",
    )
    effect_parser.add_argument(
        "--row-gap",
        default=5,
        type=argtypes.valid_gap,
        help="Number of characters to wait before adding a new row. Min 1.",
    )
    effect_parser.add_argument(
        "--slide-direction",
        default="left",
        choices=["left", "right"],
        help="Direction the text will slide.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.8,
        metavar="(float > 0)",
        help="Movement speed of the characters. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="IN_OUT_QUAD",
        type=argtypes.valid_ease,
        help="Easing function to use for row movement.",
    )


class SlideDirection(Enum):
    LEFT = auto()
    RIGHT = auto()


class RowSlide(base_effect.Effect):
    """Effect that slides each row into place."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        """Effect that slides each row into place.

        Args:
            terminal (Terminal): terminal to use for the effect
            args (argparse.Namespace): arguments from argparse
        """
        super().__init__(terminal, args)
        self.row_gap: int = args.row_gap
        if args.slide_direction == "left":
            self.slide_direction = SlideDirection.LEFT
        else:
            self.slide_direction = SlideDirection.RIGHT

    def prepare_data(self) -> None:
        """Prepares the data for the effect by grouping the characters by row and setting the starting
        coordinate."""

        self.rows = self.input_by_row()
        for row in self.rows.values():
            for character in row:
                if self.slide_direction == SlideDirection.LEFT:
                    character.motion.set_coordinate(self.terminal.output_area.right, character.input_coord.row)
                else:
                    character.motion.set_coordinate(0, character.input_coord.row)
                character.motion.new_waypoint(
                    "input_coord",
                    character.input_coord.column,
                    character.input_coord.row,
                    speed=self.args.movement_speed,
                    ease=self.args.easing,
                )
                character.motion.activate_waypoint("input_coord")

    def get_next_row(self) -> list[base_character.EffectCharacter]:
        """Gets the next row of characters to animate.

        Returns:
            list[effect_char.EffectCharacter]: The next row of characters to animate.
        """
        next_row = self.rows.pop(min(self.rows.keys()))
        return next_row

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        active_rows: list[list[base_character.EffectCharacter]] = []
        active_rows.append(self.get_next_row())
        row_delay_countdown = self.row_gap
        while active_rows or self.animating_chars:
            if (row_delay_countdown == 0 and self.rows) or (not active_rows and self.rows):
                active_rows.append(self.get_next_row())
                row_delay_countdown = self.row_gap
            else:
                if self.rows:
                    row_delay_countdown -= 1
            for row in active_rows:
                if row:
                    if self.slide_direction == SlideDirection.LEFT:
                        next_character = row.pop(0)
                        next_character.is_active = True
                        self.animating_chars.append(next_character)
                    else:
                        next_character = row.pop(-1)
                        next_character.is_active = True
                        self.animating_chars.append(next_character)
            self.animate_chars()
            self.terminal.print()
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.motion.movement_complete()
            ]
            active_rows = [row for row in active_rows if row]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()



import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects import base_character, base_effect
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, argtypes


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "fireworks",
        formatter_class=argtypes.CustomFormatter,
        help="Characters launch and explode like fireworks and fall into place.",
        description="fireworks | Characters explode like fireworks and fall into place.",
        epilog="Example: terminaltexteffects fireworks -a 0.01 --firework-colors 88F7E2 44D492 F5EB67 FFA15C FA233E --final-color 000000 --launch-delay 30 --launch-easing OUT_EXPO --launch-speed 0.2 --explode-easing OUT_QUAD --explode-speed 0.3 --fall-easing IN_SINE --fall-speed 0.6",
    )
    effect_parser.set_defaults(effect_class=FireworksEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time, in seconds, between animation steps.",
    )
    effect_parser.add_argument(
        "--explode-anywhere",
        action="store_true",
        default=False,
        help="If set, fireworks explode anywhere in the output area. Otherwise, fireworks explode above highest settled row of text.",
    )
    effect_parser.add_argument(
        "--firework-colors",
        type=argtypes.valid_color,
        nargs="*",
        default=0,
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Space separated list of colors from which firework colors will be randomly selected.",
    )
    effect_parser.add_argument(
        "--firework-symbol",
        type=argtypes.valid_symbol,
        default="⯏",
        metavar="(single character)",
        help="Symbol to use for the firework shell. Defaults to ⯏.",
    )
    effect_parser.add_argument(
        "--firework-volume",
        type=argtypes.positive_int,
        default=8,
        metavar="(int > 0)",
        help="Number of characters in each firework shell. Defaults to 8.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the final character. Defaults to white.",
    )
    effect_parser.add_argument(
        "--launch-delay",
        type=argtypes.ge_zero,
        default=30,
        metavar="(int >= 0)",
        help="Number of animation steps to wait between launching each firework shell.",
    )
    effect_parser.add_argument(
        "--launch-easing",
        default="OUT_EXPO",
        type=argtypes.valid_ease,
        help="Easing function to use firework launch.",
    )
    effect_parser.add_argument(
        "--launch-speed",
        type=argtypes.valid_speed,
        default=0.2,
        metavar="(float > 0)",
        help="Firework launch speed. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--explode-distance",
        default=10,
        type=argtypes.positive_int,
        metavar="(int > 0)",
        help="Maximum distance a character can travel as a part of the firework explosion.",
    )
    effect_parser.add_argument(
        "--explode-easing",
        default="OUT_QUAD",
        type=argtypes.valid_ease,
        help="Easing function to use firework explosion.",
    )
    effect_parser.add_argument(
        "--explode-speed",
        type=argtypes.valid_speed,
        default=0.3,
        metavar="(float > 0)",
        help="Firework explode speed. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--fall-easing",
        default="IN_OUT_CUBIC",
        type=argtypes.valid_ease,
        help="Easing function to use for characters falling into place.",
    )
    effect_parser.add_argument(
        "--fall-speed",
        type=argtypes.valid_speed,
        default=0.4,
        metavar="(float > 0)",
        help="Character fall speed. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )


class FireworksEffect(base_effect.Effect):
    """Effect that launches characters up the screen where they explode like fireworks and fall into place."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        super().__init__(terminal, args)
        self.shells: list[list[base_character.EffectCharacter]] = []
        if self.args.firework_colors:
            self.firework_colors = self.args.firework_colors
        else:
            self.firework_colors = ["88F7E2", "44D492", "F5EB67", "FFA15C", "FA233E"]

    def prepare_waypoints(self) -> None:
        firework_shell = []
        origin_x = random.randrange(0, self.terminal.output_area.right)
        origin_y = random.randrange(0, self.terminal.output_area.top)
        for character in self.terminal.characters:
            if len(firework_shell) == self.args.firework_volume:
                self.shells.append(firework_shell)
                firework_shell = []
                origin_x = random.randrange(0, self.terminal.output_area.right)
                if not self.args.explode_anywhere:
                    min_row = character.input_coord.row
                else:
                    min_row = self.terminal.output_area.bottom
                origin_y = random.randrange(min_row, self.terminal.output_area.top + 1)
            point_on_circle = random.choice(
                character.motion.find_points_on_circle(
                    (origin_x, origin_y), random.randrange(1, self.args.explode_distance), self.args.firework_volume
                )
            )
            character.motion.set_coordinate(origin_x, self.terminal.output_area.bottom)
            character.motion.new_waypoint(
                "apex", origin_x, origin_y, speed=self.args.launch_speed, ease=self.args.launch_easing
            )
            character.motion.new_waypoint(
                "explode",
                point_on_circle.column,
                point_on_circle.row,
                speed=self.args.explode_speed,
                ease=self.args.explode_easing,
            )
            character.motion.new_waypoint(
                "input_coord",
                character.input_coord.column,
                character.input_coord.row,
                speed=self.args.fall_speed,
                ease=self.args.fall_easing,
            )
            character.motion.activate_waypoint("apex")
            character.event_handler.register_event(
                character.event_handler.Event.WAYPOINT_REACHED,
                "apex",
                character.event_handler.Action.ACTIVATE_WAYPOINT,
                "explode",
            )
            character.event_handler.register_event(
                character.event_handler.Event.WAYPOINT_REACHED,
                "explode",
                character.event_handler.Action.ACTIVATE_WAYPOINT,
                "input_coord",
            )
            firework_shell.append(character)
        if firework_shell:
            self.shells.append(firework_shell)

    def prepare_scenes(self) -> None:
        for firework_shell in self.shells:
            shell_color = random.choice(self.firework_colors)
            for character in firework_shell:
                # launch scene
                for color in self.firework_colors:
                    character.animator.add_effect_to_scene("launch", self.args.firework_symbol, shell_color, 2)
                    character.animator.add_effect_to_scene("launch", self.args.firework_symbol, "FFFFFF", 1)
                    character.animator.scenes["launch"].is_looping = True
                # bloom scene
                bloom_gradient = graphics.gradient(shell_color, self.args.final_color, 15)
                for color in bloom_gradient:
                    character.animator.add_effect_to_scene("bloom", character.input_symbol, color, 15)
                character.animator.activate_scene("launch")
                character.event_handler.register_event(
                    character.event_handler.Event.WAYPOINT_REACHED,
                    "apex",
                    character.event_handler.Action.ACTIVATE_SCENE,
                    "bloom",
                )

    def prepare_data(self) -> None:
        """Prepares the data for the effect by ___."""
        self.prepare_waypoints()
        self.prepare_scenes()

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        launch_delay = 0
        while self.shells or self.animating_chars:
            if self.shells and launch_delay == 0:
                next_group = self.shells.pop()
                for character in next_group:
                    character.is_active = True
                    self.animating_chars.append(character)
                launch_delay = self.args.launch_delay + 1
            self.terminal.print()
            self.animate_chars()
            launch_delay -= 1

            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animator.is_active_scene_complete()
                or not animating_char.motion.movement_complete()
            ]
        self.terminal.print()

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation."""
        for animating_char in self.animating_chars:
            animating_char.animator.step_animation()
            animating_char.motion.move()



import argparse
import random
from dataclasses import dataclass

from terminaltexteffects import base_character, base_effect
from terminaltexteffects.utils import graphics, argtypes
from terminaltexteffects.utils.terminal import Terminal


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "decrypt",
        help="Display a movie style decryption effect.",
        description="decrypt | Movie style decryption effect.",
        epilog="Example: terminaltexteffects decrypt -a 0.003 --cipher-text-color 40 --plain-text-color 208",
    )
    effect_parser.set_defaults(effect_class=DecryptEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.003,
        help="Time to sleep between animation steps. Defaults to 0.003 seconds.",
    )
    effect_parser.add_argument(
        "--ciphertext-color",
        type=argtypes.valid_color,
        default=40,
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the ciphertext. Defaults to 40",
    )
    effect_parser.add_argument(
        "--plaintext-color",
        type=argtypes.valid_color,
        default="ffb007",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the plaintext. Defaults to 208.",
    )


@dataclass
class DecryptChars:
    """Various decimal utf-8 character ranges."""

    keyboard = list(range(33, 127))
    blocks = list(range(9608, 9632))
    box_drawing = list(range(9472, 9599))
    misc = list(range(174, 452))


class DecryptEffect(base_effect.Effect):
    """Effect that shows a movie style text decryption effect."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        super().__init__(terminal, args)
        self.ciphertext_color = args.ciphertext_color
        self.plaintext_color = args.plaintext_color
        self.character_discovered_gradient = graphics.gradient("ffffff", self.plaintext_color, 5)
        self.encrypted_symbols: list[str] = []
        self.make_encrypted_symbols()

    def make_encrypted_symbols(self) -> None:
        for n in DecryptChars.keyboard:
            self.encrypted_symbols.append(chr(n))
        for n in DecryptChars.blocks:
            self.encrypted_symbols.append(chr(n))
        for n in DecryptChars.box_drawing:
            self.encrypted_symbols.append(chr(n))
        for n in DecryptChars.misc:
            self.encrypted_symbols.append(chr(n))

    def make_decrypting_animation_scenes(self, character: base_character.EffectCharacter) -> None:
        for _ in range(80):
            symbol = random.choice(self.encrypted_symbols)
            duration = 3
            character.animator.add_effect_to_scene("fast_decrypt", symbol, self.ciphertext_color, duration)
        for _ in range(random.randint(1, 15)):  # 1-15 longer duration units
            symbol = random.choice(self.encrypted_symbols)
            if random.randint(0, 100) <= 30:  # 30% chance of extra long duration
                duration = random.randrange(75, 225)  # wide long duration range reduces 'waves' in the animation
            else:
                duration = random.randrange(5, 10)  # shorter duration creates flipping effect
            character.animator.add_effect_to_scene("slow_decrypt", symbol, self.ciphertext_color, duration)
        for color in self.character_discovered_gradient:
            character.animator.add_effect_to_scene("discovered", character.input_symbol, color, 20)

        character.animator.add_effect_to_scene("discovered", character.input_symbol, self.plaintext_color, 1)

    def prepare_data_for_type_effect(self) -> None:
        """Prepares the data for the effect by building the animation for each character."""
        for character in self.terminal.characters:
            for block_char in ["▉", "▓", "▒", "░"]:
                character.animator.add_effect_to_scene("typing", block_char, self.ciphertext_color, 2)
            character.animator.add_effect_to_scene(
                "typing", random.choice(self.encrypted_symbols), self.ciphertext_color, 2
            )
            self.pending_chars.append(character)

    def prepare_data_for_decrypt_effect(self) -> None:
        """Prepares the data for the effect by building the animation for each character."""
        for character in self.terminal.characters:
            self.make_decrypting_animation_scenes(character)
            character.event_handler.register_event(
                character.event_handler.Event.SCENE_COMPLETE,
                "fast_decrypt",
                character.event_handler.Action.ACTIVATE_SCENE,
                "slow_decrypt",
            )
            character.event_handler.register_event(
                character.event_handler.Event.SCENE_COMPLETE,
                "slow_decrypt",
                character.event_handler.Action.ACTIVATE_SCENE,
                "discovered",
            )
            character.animator.activate_scene("fast_decrypt")
            self.animating_chars.append(character)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data_for_type_effect()
        self.run_type_effect()
        self.prepare_data_for_decrypt_effect()
        self.run_decryption_effect()

    def run_type_effect(self) -> None:
        """Runs the typing out the characters effect."""
        self.terminal.print()
        while self.pending_chars or self.animating_chars:
            if self.pending_chars:
                if random.randint(0, 100) <= 75:
                    next_character = self.pending_chars.pop(0)
                    next_character.is_active = True
                    next_character.animator.activate_scene("typing")
                    self.animating_chars.append(next_character)
            self.animate_chars()
            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animator.is_active_scene_complete()
            ]
            self.terminal.print()

    def run_decryption_effect(self) -> None:
        while self.animating_chars:
            self.animate_chars()
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animator.is_active_scene_complete()
            ]
            self.terminal.print()

    def animate_chars(self) -> None:
        """Animates the characters by calling the tween method and printing the characters to the terminal."""
        for animating_char in self.animating_chars:
            animating_char.animator.step_animation()



import argparse
import importlib
import pkgutil

import terminaltexteffects.effects
import terminaltexteffects.utils.terminal as term


def main():
    parser = (argparse.ArgumentParser)(
        prog="terminaltexteffects",
        description="Apply visual effects to terminal text piped in from stdin.",
        epilog="Ex: ls -a | python -m terminaltexteffects --xterm-colors decrypt -a 0.002 --ciphertext-color 00ff00 --plaintext-color ff0000 --final-color 0000ff",
    )
    parser.add_argument(
        "--xterm-colors",
        action="store_true",
        help="Convert any colors specified in RBG hex to the closest XTerm-256 color.",
        default=False,
    )
    parser.add_argument(
        "--no-color",
        action="store_true",
        help="Disable all colors in the effect.",
        default=False,
    )
    subparsers = parser.add_subparsers(
        title="Effect",
        description="Name of the effect to apply. Use <effect> -h for effect specific help.",
        help="Available Effects",
        required=True,
    )

    discovered_effects = [
        importlib.import_module(module.name)
        for module in pkgutil.iter_modules(
            terminaltexteffects.effects.__path__, terminaltexteffects.effects.__name__ + "."
        )
    ]

    for effect in discovered_effects:
        effect.add_arguments(subparsers)

    args = parser.parse_args()
    input_data = term.Terminal.get_piped_input()
    if not input_data.strip():
        print("NO INPUT.")
    else:
        terminal = term.Terminal(input_data, args)
        effect = args.effect_class(terminal, args)
        effect.run()


if __name__ == "__main__":
    main()
